import argparse
import json
import re
import yaml
from datetime import date
from difflib import unified_diff, HtmlDiff
from typing import Union, Optional, Iterator, List, Dict, Any, NamedTuple, Type, Tuple, Callable, Set
from pydantic import BaseModel, ValidationError, validator, Extra, root_validator
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import Rest
from cisco_sdwan.base.catalog import CATALOG_TAG_ALL, ordered_tags
from cisco_sdwan.tasks.utils import TaskOptions, existing_workdir_type, filename_type, existing_file_type
from cisco_sdwan.tasks.common import Task, Table, TaskException
from cisco_sdwan.tasks.models import TaskArgs, const
from cisco_sdwan.tasks.validators import validate_existing_file, validate_filename, validate_workdir, validate_json
from ._list import TaskList, ListConfigArgs, ListCertificateArgs
from ._show_template import TaskShowTemplate, ShowTemplateValuesArgs, ShowTemplateRefArgs
from ._show import TaskShow, ShowDevicesArgs, ShowRealtimeArgs, ShowStateArgs, ShowStatisticsArgs


# Models for the report specification
class SectionModel(BaseModel, extra=Extra.forbid):
    name: str
    task: str
    args: Optional[Dict[str, Any]] = None
    inherit_globals: bool = True
    skip_diff: bool = False

    @validator('task')
    def validate_task(cls, v):
        if tuple(v.split()) not in section_catalog:
            raise ValueError(f"'{v}' is not a valid section task")
        return v

    @property
    def task_label(self) -> Tuple[str, ...]:
        return tuple(self.task.split())


class ReportContentModel(BaseModel):
    globals: Optional[Dict[str, Any]] = None
    sections: List[SectionModel]

    @property
    def skip_section_set(self) -> Set[str]:
        return {section.name for section in self.sections if section.skip_diff}


# Report specification used as default if user did not provide a custom one
DEFAULT_SECTIONS_1: List[Dict[str, Any]] = [
    {'name': f'List configuration {tag}', 'task': 'list configuration', 'args': {'tags': [tag]}}
    for tag in ordered_tags(CATALOG_TAG_ALL)
]
DEFAULT_CONTENT_SPEC = {
    'sections': DEFAULT_SECTIONS_1 + [
        {'name': 'List certificate', 'task': 'list certificate'},
        {'name': 'Show-template values', 'task': 'show-template values'},
        {'name': 'Show-template references', 'task': 'show-template references'},
        {'name': 'Show devices', 'task': 'show devices'},
        {'name': 'Show state', 'task': 'show state', 'args': {'cmd': ['all']}},
    ]
}


# Metadata about tasks that can be included in a report
class TaskMeta(NamedTuple):
    task_cls: Type[Task]
    task_args_cls: Type[TaskArgs]


section_catalog: Dict[Tuple[str, ...], TaskMeta] = {
    ('show', 'devices'): TaskMeta(TaskShow, ShowDevicesArgs),
    ('show', 'realtime'): TaskMeta(TaskShow, ShowRealtimeArgs),
    ('show', 'state'): TaskMeta(TaskShow, ShowStateArgs),
    ('show', 'statistics'): TaskMeta(TaskShow, ShowStatisticsArgs),
    ('list', 'certificate'): TaskMeta(TaskList, ListCertificateArgs),
    ('list', 'configuration'): TaskMeta(TaskList, ListConfigArgs),
    ('show-template', 'values'): TaskMeta(TaskShowTemplate, ShowTemplateValuesArgs),
    ('show-template', 'references'): TaskMeta(TaskShowTemplate, ShowTemplateRefArgs)
}


def load_content_spec(spec_file: Optional[str], spec_json: Optional[str],
                      spec_default: Optional[dict] = None) -> ReportContentModel:
    def load_yaml(filename):
        try:
            with open(filename) as yaml_file:
                return yaml.safe_load(yaml_file)
        except FileNotFoundError as ex:
            raise FileNotFoundError(f'Could not load report specification file: {ex}') from None
        except yaml.YAMLError as ex:
            raise TaskException(f'Report specification YAML syntax error: {ex}') from None

    def load_json(json_str):
        try:
            return json.loads(json_str)
        except json.JSONDecodeError as ex:
            raise TaskException(f'Report specification JSON syntax error: {ex}') from None

    if spec_file:
        content_spec_dict = load_yaml(spec_file)
    elif spec_json:
        content_spec_dict = load_json(spec_json)
    else:
        content_spec_dict = spec_default

    try:
        return ReportContentModel(**content_spec_dict)
    except ValidationError as e:
        raise TaskException(f'Invalid report specification: {e}') from None


class Report:
    DEFAULT_SUBSECTION_NAME = "Default"

    def __init__(self, filename: str, section_dict: Optional[dict] = None) -> None:
        # Report section_dict is {<section_name>: {<subsection_name>: [<subsection lines]}}
        self.section_dict = {} if section_dict is None else section_dict
        self.filename = filename

    def add_section(self, section_name: str, subsection_list: list) -> None:
        for subsection in subsection_list:
            if isinstance(subsection, Table):
                subsection_name = subsection.name or Report.DEFAULT_SUBSECTION_NAME
                subsection_lines = list(subsection.pretty_iter())
            else:
                subsection_name = Report.DEFAULT_SUBSECTION_NAME
                subsection_lines = [str(subsection)]

            self.section_dict.setdefault(section_name, {}).setdefault(subsection_name, []).extend(subsection_lines)

    def render(self) -> Iterator[str]:
        for section_num, (section_name, subsection_dict) in enumerate(self.section_dict.items()):
            if section_num != 0:
                yield ''

            yield f"### {section_name} ###"
            for subsection in subsection_dict.values():
                yield ''
                yield from subsection
            yield ''

    def __str__(self) -> str:
        return '\n'.join(self.render())

    def save(self) -> None:
        with open(self.filename, 'w') as f:
            f.write(str(self))

    @classmethod
    def load(cls, filename: str):
        try:
            with open(filename) as report_f:
                report_data = report_f.read()
        except FileNotFoundError as ex:
            raise FileNotFoundError(f"Failed to load report file: {ex}") from None

        p_section_name = re.compile(r"###(?P<section_name>[^#]+)###$")
        p_subsection_name = re.compile(r"\*{3}(?P<subsection_name>[^*]+)\*{3}$")

        section_dict = {}
        section_name, subsection_name = None, None
        for line in report_data.splitlines():
            if not line.strip():
                continue

            m_section_name = p_section_name.match(line)
            if m_section_name:
                section_name = m_section_name.group('section_name').strip()
                subsection_name = Report.DEFAULT_SUBSECTION_NAME
                continue

            if section_name is None or subsection_name is None:
                continue

            m_subsection_name = p_subsection_name.match(line)
            if m_subsection_name:
                subsection_name = m_subsection_name.group('subsection_name').strip()

            section_dict.setdefault(section_name, {}).setdefault(subsection_name, []).append(line)

        return cls(filename, section_dict)

    def trimmed(self, trim_section_names: set):
        trimmed_section_dict = {
            section_name: section for section_name, section in self.section_dict.items()
            if section_name not in trim_section_names
        }
        return Report(self.filename, trimmed_section_dict)


@TaskOptions.register('report')
class TaskReport(Task):
    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nReport task:')
        task_parser.prog = f'{task_parser.prog} report'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        sub_tasks = task_parser.add_subparsers(title='report options')
        sub_tasks.required = True

        create_parser = sub_tasks.add_parser('create', help='create a report')
        create_parser.set_defaults(subtask_handler=TaskReport.subtask_create)
        create_parser.add_argument('--file', metavar='<filename>', type=filename_type,
                                   default=f'report_{date.today():%Y%m%d}.txt',
                                   help='report filename (default: %(default)s)')
        create_parser.add_argument('--workdir', metavar='<directory>', type=existing_workdir_type,
                                   help='report from the specified directory instead of target vManage')
        create_parser.add_argument('--diff', metavar='<filename>', type=existing_file_type,
                                   help='generate diff between the specified previous report and the current report')

        diff_parser = sub_tasks.add_parser('diff', help='generate diff between two reports')
        diff_parser.set_defaults(subtask_handler=TaskReport.subtask_diff)
        diff_parser.add_argument('report_a', metavar='<report a>', type=existing_file_type,
                                 help='report a filename (from)')
        diff_parser.add_argument('report_b', metavar='<report b>', type=existing_file_type,
                                 help='report b filename (to)')
        diff_parser.add_argument('--save-html', metavar='<filename>', type=filename_type,
                                 help='save report diff as html file')
        diff_parser.add_argument('--save-txt', metavar='<filename>', type=filename_type,
                                 help='save report diff as text file')

        for sub_task in (create_parser, diff_parser):
            mutex = sub_task.add_mutually_exclusive_group()
            mutex.add_argument('--spec-file', metavar='<filename>', type=existing_file_type,
                               help='load custom report specification from YAML file')
            mutex.add_argument('--spec-json', metavar='<json>',
                               help='load custom report specification from JSON-formatted string')

        return task_parser.parse_args(task_args)

    @staticmethod
    def is_api_required(parsed_args) -> bool:
        return parsed_args.subtask_handler is TaskReport.subtask_create and parsed_args.workdir is None

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        return parsed_args.subtask_handler(self, parsed_args, api)

    def subtask_create(self, parsed_args, api: Optional[Rest]) -> Union[None, list]:
        source_info = f'Local workdir: "{parsed_args.workdir}"' if api is None else f'vManage URL: "{api.base_url}"'
        self.log_info(f'Report create task: {source_info} -> "{parsed_args.file}"')

        self.log_info("Loading report specification")
        content_spec = load_content_spec(parsed_args.spec_file, parsed_args.spec_json, DEFAULT_CONTENT_SPEC)

        report = Report(parsed_args.file)
        for description, task_cls, task_args in self.section_iter(content_spec, api is not None, parsed_args.workdir):
            try:
                task_output = task_cls().runner(task_args, api)
                if task_output:
                    report.add_section(description, task_output)
            except (TaskException, FileNotFoundError) as ex:
                self.log_error(f'Task {task_cls.__name__} error: {ex}')

        result = None
        if parsed_args.diff:
            self.log_info(f'Starting diff from "{parsed_args.diff}" to "{parsed_args.file}"')
            previous_report = Report.load(parsed_args.diff)
            self.log_info(f'Loaded previous report "{parsed_args.diff}"')
            skip_sections = content_spec.skip_section_set
            result = [diff_txt(previous_report.trimmed(skip_sections), report.trimmed(skip_sections))]
            self.log_info('Completed diff')

        # Saving current report after running the diff in case the previous report had the same filename
        report.save()
        self.log_info(f'Report saved as "{parsed_args.file}"')

        return result

    def subtask_diff(self, parsed_args, api: Optional[Rest]) -> Union[None, list]:
        self.log_info(f'Report diff task: "{parsed_args.report_a}" <-> "{parsed_args.report_b}"')
        report_a = Report.load(parsed_args.report_a)
        self.log_info(f'Loaded report "{parsed_args.report_a}"')
        report_b = Report.load(parsed_args.report_b)
        self.log_info(f'Loaded report "{parsed_args.report_b}"')

        if parsed_args.spec_file or parsed_args.spec_json:
            self.log_info("Loading report specification, trimming sections with skip_diff set")
            skip_sections = load_content_spec(parsed_args.spec_file, parsed_args.spec_json).skip_section_set
            report_a = report_a.trimmed(skip_sections)
            report_b = report_b.trimmed(skip_sections)

        result = None
        if parsed_args.save_html:
            with open(parsed_args.save_html, 'w') as f:
                f.write(diff_html(report_a, report_b))
            self.log_info(f'HTML report diff saved as "{parsed_args.save_html}"')

        if parsed_args.save_txt:
            with open(parsed_args.save_txt, 'w') as f:
                f.write(diff_txt(report_a, report_b))
            self.log_info(f'Text report diff saved as "{parsed_args.save_txt}"')

        if not parsed_args.save_html and not parsed_args.save_txt:
            result = [diff_txt(report_a, report_b)]

        return result

    def section_iter(self, report_spec: ReportContentModel,
                     has_api_session: bool,
                     workdir: Optional[str]) -> Iterator[Tuple[str, Type[Task], TaskArgs]]:
        """
        An iterator over the different sections of the report, including task and arguments for the task.
        @param report_spec: report specification to use for generating this report
        @param has_api_session: whether the report is running with a vManage session or offline (i.e. off a backup)
        @param workdir: workdir value, if provided to the report task.
        @return: an iterator of (<description>, <task class>, <task args>)
        """
        spec_global_args = report_spec.globals or {}

        for section_num, section in enumerate(report_spec.sections):
            task_meta: TaskMeta = section_catalog[section.task_label]

            # Resolving task args: section args have higher precedence over global args
            # Trim global args to only the ones relevant to the task, then merge with section args
            spec_section_args = section.args or {}
            if section.inherit_globals:
                imported_global_args = {
                    k: v for k, v in spec_global_args.items() if k in task_meta.task_args_cls.__fields__
                }
                spec_args = {**imported_global_args, **spec_section_args}
            else:
                spec_args = spec_section_args

            # Parse task args
            try:
                task_args = task_meta.task_args_cls(**spec_args)
            except ValidationError as ex:
                self.log_error(f"Invalid report specification: {section.name} (sections -> {section_num}): {ex}")
                continue

            if workdir is not None and hasattr(task_args, 'workdir') and task_args.workdir is None:
                task_args.workdir = workdir

            if task_meta.task_cls.is_api_required(task_args) and not has_api_session:
                # Skip report sections that require an api session when report is run offline, i.e. from a workdir
                self.log_debug(f"Skipping: {section.name} (sections -> {section_num}): report from workdir")
                continue

            yield section.name, task_meta.task_cls, task_args


def diff_html(a: Report, b: Report) -> str:
    diff = HtmlDiff()
    return diff.make_file(list(a.render()), list(b.render()), fromdesc=a.filename, todesc=b.filename, context=False)


def diff_txt(a: Report, b: Report) -> str:
    return '\n'.join(diff_txt_iter(a, b))


def diff_txt_iter(a: Report, b: Report) -> Iterator[str]:
    def spaced_line(line: str) -> Iterator[str]:
        yield ""
        yield line

    def subsection_lines(subsection_iter: Iterator[str]) -> List[str]:
        lines = []
        for subsection in subsection_iter:
            lines.extend(subsection.splitlines())
        return lines

    yield from spaced_line(f"### Report Diff - a/{a.filename} <-> b/{b.filename} ###")

    # Find sections that were on <a> but not on <b>
    for a_section_name in set(a.section_dict) - set(b.section_dict):
        yield from spaced_line(f"deleted a/{a_section_name.lower()}")

    # Diff each section
    for section_name, b_subsection_dict in b.section_dict.items():
        a_subsection_dict = a.section_dict.get(section_name)
        if a_subsection_dict is None:
            yield from spaced_line(f"new b/{section_name.lower()}")
            continue

        # Find subsections that were on <a> but not on <b>
        for a_subsection_name in set(a_subsection_dict) - set(b_subsection_dict):
            yield from spaced_line(f"deleted a/{a_subsection_name.lower()}")

        # Diff each subsection
        for subsection_name, b_subsections in b_subsection_dict.items():
            a_subsections = a_subsection_dict.get(subsection_name)
            if a_subsections is None:
                yield from spaced_line(f"new b/{subsection_name.lower()}")
                continue

            diff_info = [section_name.lower()]
            if subsection_name != Report.DEFAULT_SUBSECTION_NAME:
                diff_info.append(subsection_name.lower())

            is_first = True
            for diff_line in unified_diff(subsection_lines(a_subsections), subsection_lines(b_subsections),
                                          fromfile=f"a/{a.filename}", tofile=f"b/{b.filename}", lineterm='', n=1):
                if is_first:
                    yield from spaced_line(f"changed {', '.join(diff_info)}")
                    is_first = False

                yield diff_line

    yield from spaced_line("### End Report Diff ###")
    yield ""


class ReportCreateArgs(TaskArgs):
    subtask_handler: Callable = const(TaskReport.subtask_create)
    file: Optional[str] = None
    workdir: Optional[str] = None
    diff: Optional[str] = None
    spec_file: Optional[str] = None
    spec_json: Optional[str] = None

    # Validators
    _validate_workdir = validator('workdir', allow_reuse=True)(validate_workdir)
    _validate_existing_file = validator('spec_file', 'diff', allow_reuse=True)(validate_existing_file)
    _validate_json = validator('spec_json', allow_reuse=True)(validate_json)

    @validator('file', pre=True, always=True)
    def validate_report_file(cls, v):
        filename = v or f'report_{date.today():%Y%m%d}.txt'
        return validate_filename(filename)

    @root_validator(skip_on_failure=True)
    def mutex_validations(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get('spec_file') is not None and values.get('spec_json') is not None:
            raise ValueError('Argument "spec_file" not allowed with "spec_json"')

        return values


class ReportDiffArgs(TaskArgs):
    subtask_handler: Callable = const(TaskReport.subtask_diff)
    report_a: Optional[str] = None
    report_b: Optional[str] = None
    save_html: Optional[str] = None
    save_txt: Optional[str] = None
    spec_file: Optional[str] = None
    spec_json: Optional[str] = None

    # Validators
    _validate_existing_file = validator('report_a', 'report_b', 'spec_file', allow_reuse=True)(validate_existing_file)
    _validate_json = validator('spec_json', allow_reuse=True)(validate_json)
    _validate_filename = validator('save_html', 'save_txt', allow_reuse=True)(validate_filename)

    @root_validator(skip_on_failure=True)
    def mutex_validations(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get('spec_file') is not None and values.get('spec_json') is not None:
            raise ValueError('Argument "spec_file" not allowed with "spec_json"')

        return values



import argparse
from typing import Union, Optional, Callable, Dict, Any
from pydantic import validator, root_validator
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import Rest, RestAPIException
from cisco_sdwan.base.models_vmanage import EdgeCertificate, EdgeCertificateSync
from cisco_sdwan.tasks.utils import TaskOptions, existing_workdir_type, regex_type, default_workdir
from cisco_sdwan.tasks.common import regex_search, Task, WaitActionsException
from cisco_sdwan.tasks.models import TaskArgs, const
from cisco_sdwan.tasks.validators import validate_regex, validate_workdir


@TaskOptions.register('certificate')
class TaskCertificate(Task):
    SAVINGS_FACTOR = 0.2

    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nCertificate task:')
        task_parser.prog = f'{task_parser.prog} certificate'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        sub_tasks = task_parser.add_subparsers(title='certificate options')
        sub_tasks.required = True

        restore_parser = sub_tasks.add_parser('restore', help='restore certificate status from a backup')
        restore_parser.set_defaults(source_iter=TaskCertificate.restore_iter)
        restore_parser.add_argument('--workdir', metavar='<directory>', type=existing_workdir_type,
                                    default=default_workdir(target_address),
                                    help='restore source (default: %(default)s)')

        set_parser = sub_tasks.add_parser('set', help='set certificate status')
        set_parser.set_defaults(source_iter=TaskCertificate.set_iter)
        set_parser.add_argument('status', choices=['invalid', 'staging', 'valid'],
                                help='WAN edge certificate status')

        # Parameters common to all sub-tasks
        for sub_task in (restore_parser, set_parser):
            mutex = sub_task.add_mutually_exclusive_group()
            mutex.add_argument('--regex', metavar='<regex>', type=regex_type,
                               help='regular expression selecting devices to modify certificate status. Matches on '
                                    'the hostname or chassis/uuid. Use "^-$" to match devices without a hostname.')
            mutex.add_argument('--not-regex', metavar='<regex>', type=regex_type,
                               help='regular expression selecting devices NOT to modify certificate status. Matches on '
                                    'the hostname or chassis/uuid.')
            sub_task.add_argument('--dryrun', action='store_true',
                                  help='dry-run mode. List modifications that would be performed without pushing '
                                       'changes to vManage.')

        return task_parser.parse_args(task_args)

    @staticmethod
    def restore_iter(target_certs, parsed_args):
        saved_certs = EdgeCertificate.load(parsed_args.workdir)
        if saved_certs is None:
            raise FileNotFoundError('WAN edge certificates were not found in the backup')

        saved_certs_dict = {uuid: status for uuid, status in saved_certs}

        return (
            (uuid, status, hostname, saved_certs_dict[uuid])
            for uuid, status, hostname, chassis, serial, state in target_certs.extended_iter()
            if uuid in saved_certs_dict
        )

    @staticmethod
    def set_iter(target_certs, parsed_args):
        return (
            (uuid, status, hostname, parsed_args.status)
            for uuid, status, hostname, chassis, serial, state in target_certs.extended_iter()
        )

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        self.is_dryrun = parsed_args.dryrun
        if parsed_args.source_iter is TaskCertificate.restore_iter:
            start_msg = f'Restore status from workdir: "{parsed_args.workdir}" -> vManage URL: "{api.base_url}"'
        else:
            start_msg = f'Set status to "{parsed_args.status}" -> vManage URL: "{api.base_url}"'
        self.log_info(f'Certificate task: {start_msg}')

        try:
            self.log_info('Loading WAN edge certificate list from target vManage', dryrun=False)
            target_certs = EdgeCertificate.get_raise(api)

            regex = parsed_args.regex or parsed_args.not_regex
            matched_items = (
                (uuid, current_status, hostname, new_status)
                for uuid, current_status, hostname, new_status in parsed_args.source_iter(target_certs, parsed_args)
                if regex is None or regex_search(regex, hostname or '-', uuid, inverse=parsed_args.regex is None)
            )
            update_list = []
            self.log_info('Identifying items to be pushed', dryrun=False)
            for uuid, current_status, hostname, new_status in matched_items:
                if current_status == new_status:
                    self.log_debug(f'Skipping {hostname or uuid}, no changes')
                    continue

                self.log_info(f'Will update {hostname or uuid} status: {current_status} -> {new_status}')
                update_list.append((uuid, new_status))

            if len(update_list) > 0:
                self.log_info('Send certificate status changes to vManage')
                if not self.is_dryrun:
                    api.post(target_certs.status_post_data(*update_list), EdgeCertificate.api_path.post)
                    action_worker = EdgeCertificateSync(api.post({}, EdgeCertificateSync.api_path.post))
                    self.wait_actions(api, [(action_worker, None)], 'certificate sync with controllers',
                                      raise_on_failure=True)
            else:
                self.log_info('No certificate status changes to send')

        except (RestAPIException, FileNotFoundError, WaitActionsException) as ex:
            self.log_critical(f'Failed updating WAN edge certificate status: {ex}')

        return


class CertificateArgs(TaskArgs):
    regex: Optional[str] = None
    not_regex: Optional[str] = None
    dryrun: bool = False

    # Validators
    _validate_regex = validator('regex', 'not_regex', allow_reuse=True)(validate_regex)

    @root_validator(skip_on_failure=True)
    def mutex_validations(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get('regex') is not None and values.get('not_regex') is not None:
            raise ValueError('Argument "not_regex" not allowed with "regex"')

        return values


class CertificateRestoreArgs(CertificateArgs):
    source_iter: Callable = const(TaskCertificate.restore_iter)
    workdir: str

    # Validators
    _validate_workdir = validator('workdir', allow_reuse=True)(validate_workdir)


class CertificateSetArgs(CertificateArgs):
    source_iter: Callable = const(TaskCertificate.set_iter)
    status: str

    @validator('status')
    def validate_status(cls, v):
        status_options = ('invalid', 'staging', 'valid')
        if v not in status_options:
            raise ValueError(f'"{v}" is not a valid certificate status. Options are: {", ".join(status_options)}.')
        return v



import argparse
from pathlib import Path
from collections import namedtuple
from typing import List, Union, Optional, Callable
from operator import itemgetter, attrgetter
from pydantic import validator
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import RestAPIException, Rest
from cisco_sdwan.base.catalog import catalog_iter
from cisco_sdwan.base.models_base import filename_safe
from cisco_sdwan.base.models_vmanage import (DeviceTemplate, DeviceTemplateAttached, DeviceTemplateValues,
                                             DeviceTemplateIndex, FeatureTemplate, FeatureTemplateIndex)
from cisco_sdwan.tasks.utils import TaskOptions, existing_workdir_type, filename_type, regex_type
from cisco_sdwan.tasks.common import regex_search, Task, Table, get_table_filters, filtered_tables, export_json
from cisco_sdwan.tasks.models import TableTaskArgs, const
from cisco_sdwan.tasks.validators import validate_workdir, validate_regex


@TaskOptions.register('show-template')
class TaskShowTemplate(Task):
    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nShow-template task:')
        task_parser.prog = f'{task_parser.prog} show-template'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        sub_tasks = task_parser.add_subparsers(title='show-template options')
        sub_tasks.required = True

        values_parser = sub_tasks.add_parser('values', help='show values from template attachments')
        values_parser.set_defaults(subtask_handler=TaskShowTemplate.values_table)
        values_parser.set_defaults(subtask_info='values')
        values_parser.add_argument('--templates', metavar='<regex>', type=regex_type,
                                   help='regular expression selecting device templates to inspect. '
                                        'Match on template name or ID.')

        refs_parser = sub_tasks.add_parser('references', aliases=['ref'],
                                           help='show device templates that reference a feature template')
        refs_parser.set_defaults(subtask_handler=TaskShowTemplate.references_table)
        refs_parser.set_defaults(subtask_info='references')
        refs_parser.add_argument('--with-refs', action='store_true',
                                 help='include only feature-templates with device-template references')
        refs_parser.add_argument('--templates', metavar='<regex>', type=regex_type,
                                 help='regular expression matching feature template names to include')

        # Parameters common to all sub-tasks
        for sub_task in (values_parser, refs_parser):
            sub_task.add_argument('--workdir', metavar='<directory>', type=existing_workdir_type,
                                  help='show-template will read from the specified directory instead of target vManage')
            sub_task.add_argument('--exclude', metavar='<regex>', type=regex_type,
                                  help='exclude table rows matching the regular expression')
            sub_task.add_argument('--include', metavar='<regex>', type=regex_type,
                                  help='include table rows matching the regular expression, exclude all other rows')
            sub_task.add_argument('--save-csv', metavar='<directory>', type=filename_type,
                                  help='export tables as CSV files under the specified directory')
            sub_task.add_argument('--save-json', metavar='<filename>', type=filename_type,
                                  help='export results as JSON-formatted file')

        return task_parser.parse_args(task_args)

    @staticmethod
    def is_api_required(parsed_args) -> bool:
        return parsed_args.workdir is None

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        source_info = f'Local workdir: "{parsed_args.workdir}"' if api is None else f'vManage URL: "{api.base_url}"'
        self.log_info(f'Show-template {parsed_args.subtask_info} task: {source_info}')

        filters = get_table_filters(exclude_regex=parsed_args.exclude, include_regex=parsed_args.include)
        result_tables = filtered_tables(parsed_args.subtask_handler(self, parsed_args, api), *filters)

        if not result_tables:
            self.log_warning('No results found')
            return

        if parsed_args.save_csv is not None:
            Path(parsed_args.save_csv).mkdir(parents=True, exist_ok=True)
            for table in result_tables:
                table.save(Path(parsed_args.save_csv, table.meta))
            self.log_info(f"Tables exported as CSV files under directory '{parsed_args.save_csv}'")

        if parsed_args.save_json is not None:
            export_json(result_tables, parsed_args.save_json)
            self.log_info(f"Tables exported as JSON file '{parsed_args.save_json}'")

        return result_tables if (parsed_args.save_csv is None and parsed_args.save_json is None) else None

    def values_table(self, parsed_args, api: Optional[Rest]) -> List[Table]:
        def template_values(ext_name: bool, template_name: str, template_id: str) -> Union[DeviceTemplateValues, None]:
            if api is None:
                # Load from local backup
                values = DeviceTemplateValues.load(parsed_args.workdir, ext_name, template_name, template_id)
                if values is None:
                    self.log_debug(f'Skipped {template_name}. No template values file found.')
            else:
                # Load from vManage via API
                devices_attached = DeviceTemplateAttached.get(api, template_id)
                if devices_attached is None:
                    self.log_error(f'Failed to retrieve {template_name} attached devices')
                    return None

                try:
                    uuid_list = [uuid for uuid, _ in devices_attached]
                    values = DeviceTemplateValues(api.post(DeviceTemplateValues.api_params(template_id, uuid_list),
                                                           DeviceTemplateValues.api_path.post))
                except RestAPIException:
                    self.log_error(f'Failed to retrieve {template_name} values')
                    return None

            return values

        # Templates are sorted by template name then ID. Then for each template with attachments, devices are sorted
        # by name then UUID. The values for each device are sorted by the variable name
        result_tables = []
        backend = api or parsed_args.workdir
        matched_templates = [
            (item_id, item_name, index.need_extended_name, tag, info)
            for tag, info, index, item_cls in self.index_iter(backend, catalog_iter('template_device'))
            for item_id, item_name in index
            if (issubclass(item_cls, DeviceTemplate) and
                (parsed_args.templates is None or regex_search(parsed_args.templates, item_name, item_id)))
        ]
        matched_templates.sort(key=itemgetter(1, 0))
        for item_id, item_name, use_ext_name, tag, info in matched_templates:
            attached_values = template_values(use_ext_name, item_name, item_id)
            if attached_values is None:
                continue

            self.log_info(f'Inspecting {info} {item_name} values')
            var_names = attached_values.title_dict()
            for csv_id, csv_name, entry in sorted(attached_values, key=itemgetter(1, 0)):
                table = Table('Name', 'Value', 'Variable',
                              name=f"Template {item_name}, device {csv_name or csv_id}",
                              meta=f"template_values_{filename_safe(item_name, lower=True)}_{csv_name or csv_id}.csv")
                table.extend(
                    (var_names.get(var, '<not found>'), value, var)
                    for var, value in sorted(entry.items(), key=itemgetter(0))
                )
                if table:
                    result_tables.append(table)

        return result_tables

    def references_table(self, parsed_args, api: Optional[Rest]) -> List[Table]:
        FeatureInfo = namedtuple('FeatureInfo', ['name', 'type', 'attached', 'device_templates'])

        backend = api or parsed_args.workdir
        self.log_info('Inspecting feature templates')
        feature_index = self.index_get(FeatureTemplateIndex, backend)
        feature_dict = {}
        for item_id, item_name in feature_index:
            feature = self.item_get(FeatureTemplate, backend, item_id, item_name, feature_index.need_extended_name)
            if feature is None:
                self.log_error(f'Failed to load feature template {item_name}')
                continue

            feature_dict[item_id] = FeatureInfo(item_name, feature.type, feature.devices_attached, set())

        self.log_info('Inspecting device templates')
        device_index = self.index_get(DeviceTemplateIndex, backend)
        for item_id, item_name in device_index:
            device = self.item_get(DeviceTemplate, backend, item_id, item_name, device_index.need_extended_name)
            if device is None:
                self.log_error(f'Failed to load device template {item_name}')
                continue
            if device.is_type_cli:
                continue

            for feature_id in device.feature_templates:
                feature_info = feature_dict.get(feature_id)
                if feature_info is None:
                    self.log_warning(f'Template {item_name} references a missing feature template: {feature_id}')
                    continue

                feature_info.device_templates.add(item_name)

        self.log_info('Creating references table')
        # Ordered by feature template name. Then device templates are sorted by template name.
        table = Table('Feature Template', 'Type', 'Devices Attached', 'Device Templates',
                      meta="template_references.csv")
        matched_feature_templates = [
            info for info in feature_dict.values()
            if parsed_args.templates is None or regex_search(parsed_args.templates, info.name)
        ]
        matched_feature_templates.sort(key=attrgetter('name'))
        for feature_info in matched_feature_templates:
            if not parsed_args.with_refs and not feature_info.device_templates:
                table.add_marker()
                table.add(feature_info.name, feature_info.type, str(feature_info.attached), '')
                continue

            is_first = True
            for device_template in sorted(feature_info.device_templates):
                if is_first:
                    table.add_marker()
                    table.add(feature_info.name, feature_info.type, str(feature_info.attached), device_template)
                    is_first = False
                else:
                    table.add('', '', '', device_template)

        result_tables = []
        if table:
            result_tables.append(table)

        return result_tables


class ShowTemplateArgs(TableTaskArgs):
    subtask_info: str
    subtask_handler: Callable
    workdir: Optional[str] = None
    templates: Optional[str] = None

    # Validators
    _validate_workdir = validator('workdir', allow_reuse=True)(validate_workdir)
    _validate_templates = validator('templates', allow_reuse=True)(validate_regex)


class ShowTemplateValuesArgs(ShowTemplateArgs):
    subtask_info: str = const('values')
    subtask_handler: Callable = const(TaskShowTemplate.values_table)


class ShowTemplateRefArgs(ShowTemplateArgs):
    subtask_info: str = const('references')
    subtask_handler: Callable = const(TaskShowTemplate.references_table)
    with_refs: bool = False



import argparse
from typing import Union, Optional, List, Callable, Dict, Any
from operator import itemgetter
from pydantic import validator, root_validator
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import Rest
from cisco_sdwan.base.catalog import catalog_iter, CATALOG_TAG_ALL
from cisco_sdwan.base.models_base import ExtendedTemplate
from cisco_sdwan.base.models_vmanage import EdgeCertificate
from cisco_sdwan.tasks.models import TableTaskArgs, validate_catalog_tag, const
from cisco_sdwan.tasks.utils import (TaskOptions, TagOptions, existing_workdir_type, filename_type, regex_type,
                                     ext_template_type)
from cisco_sdwan.tasks.common import regex_search, Task, Table, get_table_filters, export_json
from cisco_sdwan.tasks.validators import validate_workdir, validate_ext_template, validate_regex


@TaskOptions.register('list')
class TaskList(Task):
    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nList task:')
        task_parser.prog = f'{task_parser.prog} list'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        sub_tasks = task_parser.add_subparsers(title='list options')
        sub_tasks.required = True

        config_parser = sub_tasks.add_parser('configuration', aliases=['config'], help='list configuration items')
        config_parser.set_defaults(subtask_handler=TaskList.config_table)
        config_parser.set_defaults(subtask_info='configuration')
        config_parser.add_argument('tags', metavar='<tag>', nargs='+', type=TagOptions.tag,
                                   help='one or more tags for selecting groups of items. Multiple tags should be '
                                        f'separated by space. Available tags: {TagOptions.options()}. Special tag '
                                        f'"{CATALOG_TAG_ALL}" selects all items.')

        cert_parser = sub_tasks.add_parser('certificate', aliases=['cert'], help='list device certificate information')
        cert_parser.set_defaults(subtask_handler=TaskList.cert_table)
        cert_parser.set_defaults(subtask_info='certificate')

        xform_parser = sub_tasks.add_parser('transform',
                                            help='list name transformations performed by a name-regex against '
                                                 'existing item names')
        xform_parser.set_defaults(subtask_handler=TaskList.xform_table)
        xform_parser.set_defaults(subtask_info='transform')
        xform_parser.add_argument('tags', metavar='<tag>', nargs='+', type=TagOptions.tag,
                                  help='one or more tags for selecting groups of items. Multiple tags should be '
                                       f'separated by space. Available tags: {TagOptions.options()}. Special tag '
                                       f'"{CATALOG_TAG_ALL}" selects all items.')
        xform_parser.add_argument('name_regex', metavar='<name-regex>', type=ext_template_type,
                                  help='name-regex used to transform an existing item name. Variable {name} is '
                                       'replaced with the original template name. Sections of the original template '
                                       'name can be selected using the {name <regex>} format. Where <regex> is a '
                                       'regular expression that must contain at least one capturing group. Capturing '
                                       'groups identify sections of the original name to keep.')
        xform_mutex = xform_parser.add_mutually_exclusive_group()
        xform_mutex.add_argument('--regex', metavar='<regex>', type=regex_type,
                                 help='select items matching regular expression on their original names')
        xform_mutex.add_argument('--not-regex', metavar='<regex>', type=regex_type,
                                 help='select items NOT matching regular expression on their original item names')

        # Parameters common to all sub-tasks
        for sub_task in (config_parser, cert_parser, xform_parser):
            sub_task.add_argument('--workdir', metavar='<directory>', type=existing_workdir_type,
                                  help='list will read from the specified directory instead of target vManage')
            sub_task.add_argument('--exclude', metavar='<regex>', type=regex_type,
                                  help='exclude table rows matching the regular expression')
            sub_task.add_argument('--include', metavar='<regex>', type=regex_type,
                                  help='include table rows matching the regular expression, exclude all other rows')
            sub_task.add_argument('--save-csv', metavar='<filename>', type=filename_type,
                                  help='export table as CSV-formatted file')
            sub_task.add_argument('--save-json', metavar='<filename>', type=filename_type,
                                  help='export table as JSON-formatted file')

        return task_parser.parse_args(task_args)

    @staticmethod
    def is_api_required(parsed_args) -> bool:
        return parsed_args.workdir is None

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        source_info = f'Local workdir: "{parsed_args.workdir}"' if api is None else f'vManage URL: "{api.base_url}"'
        self.log_info(f'List {parsed_args.subtask_info} task: {source_info}')

        result_table: Table = parsed_args.subtask_handler(self, parsed_args, api)

        filters = get_table_filters(exclude_regex=parsed_args.exclude, include_regex=parsed_args.include)
        result_table = result_table.filtered(*filters)
        self.log_info(f'Selection matched {len(result_table)} items')

        if not result_table:
            return

        if parsed_args.save_csv is not None:
            result_table.save(parsed_args.save_csv)
            self.log_info(f"Table exported as CSV file '{parsed_args.save_csv}'")

        if parsed_args.save_json is not None:
            export_json([result_table], parsed_args.save_json)
            self.log_info(f"Table exported as JSON file '{parsed_args.save_json}'")

        return [result_table] if (parsed_args.save_csv is None and parsed_args.save_json is None) else None

    def config_table(self, parsed_args, api: Optional[Rest]) -> Table:
        self.log_debug("Starting configuration subtask")
        backend = api or parsed_args.workdir
        # Only perform version-based filtering if backend is api
        version = None if api is None else api.server_version

        # Within each tag, table entries are sorted by item_name then item_id. Tag order is defined by the catalog.
        table = Table('Name', 'ID', 'Tag', 'Type')
        table.extend(
            (item_name, item_id, tag, info)
            for tag, info, index, item_cls in self.index_iter(backend, catalog_iter(*parsed_args.tags, version=version))
            for item_id, item_name in sorted(index, key=itemgetter(1, 0))
        )

        return table

    def cert_table(self, parsed_args, api: Optional[Rest]) -> Table:
        self.log_debug("Starting certificate subtask")
        if api is None:
            certs = EdgeCertificate.load(parsed_args.workdir)
            if certs is None:
                raise FileNotFoundError('WAN edge certificates were not found in the backup')
        else:
            certs = EdgeCertificate.get_raise(api)

        # Table will be sorted by hostname then chassis
        table = Table('Hostname', 'Chassis', 'Serial', 'State', 'Status')
        matched_items = [
            (hostname, chassis, serial, EdgeCertificate.state_str(state), status)
            for uuid, status, hostname, chassis, serial, state in certs.extended_iter(default='-')
        ]
        matched_items.sort(key=itemgetter(0, 1))
        table.extend(matched_items)

        return table

    def xform_table(self, parsed_args, api: Optional[Rest]) -> Table:
        self.log_debug("Starting transform subtask")
        backend = api or parsed_args.workdir
        # Only perform version-based filtering if backend is api
        version = None if api is None else api.server_version

        name_regex = ExtendedTemplate(parsed_args.name_regex)

        # Within a given tag, table entries are sorted by item_name. Tag order is defined by the catalog
        table = Table('Name', 'Transformed', 'Tag', 'Type')
        regex = parsed_args.regex or parsed_args.not_regex
        matched_items = [
            (item_name, name_regex(item_name), tag, info)
            for tag, info, index, item_cls in self.index_iter(backend, catalog_iter(*parsed_args.tags, version=version))
            for item_id, item_name in index
            if regex is None or regex_search(regex, item_name, inverse=parsed_args.regex is None)
        ]
        matched_items.sort(key=itemgetter(0))
        table.extend(matched_items)

        return table


class ListArgs(TableTaskArgs):
    subtask_info: str
    subtask_handler: Callable
    workdir: Optional[str] = None

    # Validators
    _validate_workdir = validator('workdir', allow_reuse=True)(validate_workdir)


class ListConfigArgs(ListArgs):
    subtask_info: str = const('configuration')
    subtask_handler: Callable = const(TaskList.config_table)
    tags: List[str]

    # Validators
    _validate_tags = validator('tags', each_item=True, allow_reuse=True)(validate_catalog_tag)


class ListCertificateArgs(ListArgs):
    subtask_info: str = const('certificate')
    subtask_handler: Callable = const(TaskList.cert_table)


class ListTransformArgs(ListArgs):
    subtask_info: str = const('transform')
    subtask_handler: Callable = const(TaskList.xform_table)
    tags: List[str]
    regex: Optional[str] = None
    not_regex: Optional[str] = None
    name_regex: str

    # Validators
    _validate_tags = validator('tags', each_item=True, allow_reuse=True)(validate_catalog_tag)
    _validate_regex = validator('regex', 'not_regex', allow_reuse=True)(validate_regex)
    _validate_name_regex = validator('name_regex', allow_reuse=True)(validate_ext_template)

    @root_validator(skip_on_failure=True)
    def mutex_validations(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get('regex') is not None and values.get('not_regex') is not None:
            raise ValueError('Argument "not_regex" not allowed with "regex"')

        return values



"""
 Sastre - Cisco-SDWAN Automation Toolset

 cisco_sdwan.tasks.implementation
 This module contains the implementation of user-facing tasks
"""
from ._backup import TaskBackup, BackupArgs
from ._restore import TaskRestore, RestoreArgs
from ._delete import TaskDelete, DeleteArgs
from ._migrate import TaskMigrate, MigrateArgs
from ._attach_detach import TaskAttach, TaskDetach, AttachVsmartArgs, AttachEdgeArgs, DetachVsmartArgs, DetachEdgeArgs
from ._certificate import TaskCertificate, CertificateSetArgs, CertificateRestoreArgs
from ._transform import TaskTransform, TransformCopyArgs, TransformRenameArgs, TransformRecipeArgs
from ._list import TaskList, ListConfigArgs, ListCertificateArgs, ListTransformArgs
from ._show_template import TaskShowTemplate, ShowTemplateValuesArgs, ShowTemplateRefArgs
from ._report import TaskReport, ReportCreateArgs, ReportDiffArgs
from ._show import (TaskShow, ShowDevicesArgs, ShowRealtimeArgs, ShowStateArgs, ShowStatisticsArgs, ShowEventsArgs,
                    ShowAlarmsArgs)

__all__ = [
    'TaskBackup',
    'TaskRestore',
    'TaskDelete',
    'TaskMigrate',
    'TaskAttach',
    'TaskDetach',
    'TaskCertificate',
    'TaskTransform',
    'TaskList',
    'TaskShowTemplate',
    'TaskReport',
    'TaskShow',
    'ReportCreateArgs',
    'ReportDiffArgs',
    'RestoreArgs',
    'MigrateArgs',
    'CertificateSetArgs',
    'CertificateRestoreArgs',
    'TransformCopyArgs',
    'TransformRenameArgs',
    'TransformRecipeArgs',
    'ListConfigArgs',
    'ListCertificateArgs',
    'ListTransformArgs',
    'ShowTemplateValuesArgs',
    'ShowTemplateRefArgs',
    'AttachVsmartArgs',
    'AttachEdgeArgs',
    'DetachVsmartArgs',
    'DetachEdgeArgs',
    'BackupArgs',
    'DeleteArgs',
    'ShowDevicesArgs',
    'ShowRealtimeArgs',
    'ShowStateArgs',
    'ShowStatisticsArgs',
    'ShowEventsArgs',
    'ShowAlarmsArgs'
]



import argparse
from typing import Tuple, List, Any, NamedTuple, Type, Union, Optional, Sequence, Callable, Dict
from pathlib import Path
from functools import partial
from concurrent import futures
from datetime import datetime, timedelta, timezone
from operator import attrgetter
from pydantic import validator, conint, root_validator
from cisco_sdwan.__version__ import __doc__ as title
from cisco_sdwan.base.rest_api import Rest
from cisco_sdwan.base.catalog import CATALOG_TAG_ALL, op_catalog_iter, OpType
from cisco_sdwan.base.models_base import (OperationalItem, RealtimeItem, BulkStatsItem, BulkStateItem, RecordItem,
                                          filename_safe)
from cisco_sdwan.base.models_vmanage import Device, Alarm, Event
from cisco_sdwan.tasks.utils import (regex_type, ipv4_type, site_id_type, filename_type, int_type, OpCmdOptions,
                                     TaskOptions, RTCmdSemantics, StateCmdSemantics, StatsCmdSemantics)
from cisco_sdwan.tasks.common import regex_search, Task, Table, get_table_filters, filtered_tables, export_json
from cisco_sdwan.tasks.models import TableTaskArgs, validate_op_cmd, const
from cisco_sdwan.tasks.validators import validate_site_id, validate_ipv4, validate_regex

THREAD_POOL_SIZE = 10
TIME_FORMAT = "%Y-%m-%dT%H:%M:%S"


class DeviceInfo(NamedTuple):
    hostname: str
    system_ip: str
    site_id: str
    state: str
    device_type: str
    model: str


def retrieve_rt_task(api_obj: Rest, rt_cls: Type[RealtimeItem], device: DeviceInfo) -> Tuple[DeviceInfo, Any]:
    return device, rt_cls.get(api_obj, device.system_ip)


def table_fields(op_cls: Type[OperationalItem], detail: bool, simple: bool) -> tuple:
    if detail and op_cls.fields_ext is not None:
        return op_cls.fields_std + op_cls.fields_ext

    if simple and op_cls.fields_sub is not None:
        return tuple(field for field in op_cls.fields_std if field not in op_cls.fields_sub)

    return op_cls.fields_std


@TaskOptions.register('show')
class TaskShow(Task):
    SAVINGS_FACTOR = 2
    STATS_AVG_INTERVAL_SECS = 300  # 5min window for statistics averages
    STATS_QUERY_RANGE_MINS = 120  # Statistics queries are from t to t - 2h

    @staticmethod
    def parser(task_args, target_address=None):
        task_parser = argparse.ArgumentParser(description=f'{title}\nShow task:')
        task_parser.prog = f'{task_parser.prog} show'
        task_parser.formatter_class = argparse.RawDescriptionHelpFormatter

        sub_tasks = task_parser.add_subparsers(title='show options')
        sub_tasks.required = True

        dev_parser = sub_tasks.add_parser('devices', aliases=['dev'], help='device list')
        dev_parser.set_defaults(subtask_handler=TaskShow.devices)
        dev_parser.set_defaults(subtask_info='devices')

        rt_parser = sub_tasks.add_parser('realtime', aliases=['rt'],
                                         help='realtime commands. Slower, but up-to-date data. vManage collect data '
                                              'from devices in realtime.')
        rt_parser.set_defaults(subtask_handler=TaskShow.realtime)
        rt_parser.set_defaults(subtask_info='realtime')

        state_parser = sub_tasks.add_parser('state', aliases=['st'],
                                            help='state commands. Faster and up-to-date synced state data.')
        state_parser.set_defaults(subtask_handler=TaskShow.bulk_state)
        state_parser.set_defaults(subtask_info='state')

        stats_parser = sub_tasks.add_parser('statistics', aliases=['stats'],
                                            help='statistics commands. Faster, but data is 30 min or more old. '
                                                 'Allows historical data queries.')
        stats_parser.set_defaults(subtask_handler=TaskShow.bulk_stats)
        stats_parser.set_defaults(subtask_info='statistics')
        stats_parser.add_argument('--days', metavar='<days>', type=partial(int_type, 0, 9999), default=0,
                                  help='query statistics from <days> ago (default: %(default)s, i.e. now)')
        stats_parser.add_argument('--hours', metavar='<hours>', type=partial(int_type, 0, 9999), default=0,
                                  help='query statistics from <hours> ago (default: %(default)s, i.e. now)')

        alarms_parser = sub_tasks.add_parser('alarms', help='display vManage alarms')
        alarms_parser.set_defaults(subtask_info='alarms')
        alarms_parser.set_defaults(subtask_op_cls=Alarm)

        events_parser = sub_tasks.add_parser('events', help='display vManage events')
        events_parser.set_defaults(subtask_info='events')
        events_parser.set_defaults(subtask_op_cls=Event)

        for sub_task, cmd_action, op_type in ((rt_parser, RTCmdSemantics, OpType.RT),
                                              (state_parser, StateCmdSemantics, OpType.STATE),
                                              (stats_parser, StatsCmdSemantics, OpType.STATS)):
            sub_task.add_argument('cmd', metavar='<cmd>', nargs='+', action=cmd_action,
                                  help='group of, or specific command to execute. '
                                       f'Group options: {OpCmdOptions.tags(op_type)}. '
                                       f'Command options: {OpCmdOptions.commands(op_type)}. '
                                       f'Group "{CATALOG_TAG_ALL}" selects all commands.')

        for sub_task in (rt_parser, state_parser, stats_parser, dev_parser):
            mutex = sub_task.add_mutually_exclusive_group()
            mutex.add_argument('--regex', metavar='<regex>', type=regex_type,
                               help='select devices matching regular expression on device name, type or model.')
            mutex.add_argument('--not-regex', metavar='<regex>', type=regex_type,
                               help='select devices NOT matching regular expression on device name, type or model.')
            sub_task.add_argument('--reachable', action='store_true', help='select devices that are reachable')
            sub_task.add_argument('--site', metavar='<id>', type=site_id_type, help='select devices with site ID')
            sub_task.add_argument('--system-ip', metavar='<ipv4>', type=ipv4_type, help='select device with system IP')

        for sub_task in (alarms_parser, events_parser):
            sub_task.set_defaults(subtask_handler=TaskShow.records)
            sub_task.add_argument('--max', metavar='<max-records>', type=partial(int_type, 1, 999999), default=100,
                                  help='maximum number records to retrieve (default: %(default)s)')
            sub_task.add_argument('--days', metavar='<days>', type=partial(int_type, 0, 9999), default=0,
                                  help='retrieve records since <days> ago (default: %(default)s)')
            sub_task.add_argument('--hours', metavar='<hours>', type=partial(int_type, 0, 9999), default=1,
                                  help='retrieve records since <hours> ago (default: %(default)s)')

        for sub_task in (rt_parser, state_parser, stats_parser, alarms_parser, events_parser):
            mutex = sub_task.add_mutually_exclusive_group()
            mutex.add_argument('--detail', action='store_true', help='detailed output (i.e. more columns)')
            mutex.add_argument('--simple', action='store_true', help='simple output (i.e. less columns)')

        for sub_task in (rt_parser, state_parser, stats_parser, dev_parser, alarms_parser, events_parser):
            sub_task.add_argument('--exclude', metavar='<regex>', type=regex_type,
                                  help='exclude table rows matching the regular expression')
            sub_task.add_argument('--include', metavar='<regex>', type=regex_type,
                                  help='include table rows matching the regular expression, exclude all other rows')
            sub_task.add_argument('--save-csv', metavar='<directory>', type=filename_type,
                                  help='export results as CSV files under the specified directory')
            sub_task.add_argument('--save-json', metavar='<filename>', type=filename_type,
                                  help='export results as JSON-formatted file')

        return task_parser.parse_args(task_args)

    def runner(self, parsed_args, api: Optional[Rest] = None) -> Union[None, list]:
        self.log_info(f'Show {parsed_args.subtask_info} task: vManage URL: "{api.base_url}"')

        filters = get_table_filters(exclude_regex=parsed_args.exclude, include_regex=parsed_args.include)
        result_tables = filtered_tables(parsed_args.subtask_handler(self, parsed_args, api), *filters)

        if not result_tables:
            return

        if parsed_args.save_csv is not None:
            Path(parsed_args.save_csv).mkdir(parents=True, exist_ok=True)
            for table in result_tables:
                filename_tokens = [parsed_args.subtask_info]
                if table.name is not None:
                    filename_tokens.append(filename_safe(table.name, lower=True).replace(' ', '_'))
                table.save(Path(parsed_args.save_csv, f"{'_'.join(filename_tokens)}.csv"))
            self.log_info(f"Tables exported as CSV files under directory '{parsed_args.save_csv}'")

        if parsed_args.save_json is not None:
            export_json(result_tables, parsed_args.save_json)
            self.log_info(f"Tables exported as JSON file '{parsed_args.save_json}'")

        return result_tables if (parsed_args.save_csv is None and parsed_args.save_json is None) else None

    def realtime(self, parsed_args, api: Rest) -> List[Table]:
        devices = self.selected_devices(parsed_args, api)
        pool_size = max(min(len(devices), THREAD_POOL_SIZE), 1)

        result_tables = []
        for info, rt_cls in op_catalog_iter(OpType.RT, *parsed_args.cmd, version=api.server_version):
            devices_in_scope = [dev_info for dev_info in devices if rt_cls.is_in_scope(dev_info.model)]
            if not devices_in_scope:
                self.log_debug(f"Skipping {info.lower()}, not applicable to any of the devices selected")
                continue

            self.log_info(f'Retrieving {info.lower()} for {len(devices_in_scope)} devices')
            with futures.ThreadPoolExecutor(pool_size) as executor:
                job_result_iter = executor.map(partial(retrieve_rt_task, api, rt_cls), devices_in_scope)

            table = None
            fields = table_fields(rt_cls, parsed_args.detail, parsed_args.simple)
            for device, rt_obj in job_result_iter:
                if rt_obj is None:
                    self.log_error(f'Failed to retrieve {info.lower()} from {device.hostname}')
                    continue

                if table is None:
                    table = Table('Device', *rt_obj.field_info(*fields), name=info)

                table.extend(
                    (device.hostname, *row_values)
                    for row_values in sorted(rt_obj.field_value_iter(*fields, **rt_cls.field_conversion_fns))
                )
                table.add_marker()

            if table:
                result_tables.append(table)

        return result_tables

    def bulk_state(self, parsed_args, api: Rest) -> List[Table]:
        devices = self.selected_devices(parsed_args, api)

        result_tables = []
        for info, op_cls in op_catalog_iter(OpType.STATE, *parsed_args.cmd, version=api.server_version):
            self.log_info(f'Retrieving {info.lower()} for {len(devices)} devices')

            op_obj: BulkStateItem = op_cls.get(api, count=10000)
            if op_obj is None:
                self.log_error(f'Failed to retrieve {info.lower()}')
                continue

            fields = table_fields(op_cls, parsed_args.detail, parsed_args.simple)
            node_data_dict = {}
            for node_id, *node_data_sample in op_obj.field_value_iter(
                    op_cls.field_node_id, *fields, **op_cls.field_conversion_fns):
                node_data_dict.setdefault(node_id, []).append(node_data_sample)

            table = self.build_table(info, op_obj.field_info(*fields), devices, node_data_dict)
            if table:
                result_tables.append(table)

        return result_tables

    def bulk_stats(self, parsed_args, api: Rest) -> List[Table]:
        devices = self.selected_devices(parsed_args, api)
        end_time = datetime.now(tz=timezone.utc) - timedelta(days=parsed_args.days, hours=parsed_args.hours)
        start_time = end_time - timedelta(minutes=self.STATS_QUERY_RANGE_MINS)
        query_params = {
            "endDate": end_time.strftime(TIME_FORMAT),
            "startDate": start_time.strftime(TIME_FORMAT),
            "count": 10000,
            "timeZone": "UTC"
        }
        self.log_info(f'Query timestamp: {end_time:%Y-%m-%d %H:%M:%S %Z}')

        result_tables = []
        for info, op_cls in op_catalog_iter(OpType.STATS, *parsed_args.cmd, version=api.server_version):
            self.log_info(f'Retrieving {info.lower()} for {len(devices)} devices')

            op_obj: BulkStatsItem = op_cls.get(api, **query_params)
            if op_obj is None:
                self.log_error(f'Failed to retrieve {info.lower()}')
                continue

            fields = table_fields(op_cls, parsed_args.detail, parsed_args.simple)
            node_data_dict = {}
            for node_id, *node_data_sample in op_obj.aggregated_value_iter(
                    self.STATS_AVG_INTERVAL_SECS, op_cls.field_node_id, *fields, **op_cls.field_conversion_fns):
                node_data_dict.setdefault(node_id, []).append(node_data_sample)

            table = self.build_table(info, op_obj.field_info(*fields), devices, node_data_dict)
            if table:
                result_tables.append(table)

        return result_tables

    def devices(self, parsed_args, api: Rest) -> List[Table]:
        devices = self.selected_devices(parsed_args, api)

        result_tables = []
        table = Table('Name', 'System IP', 'Site ID', 'Reachability', 'Type', 'Model')
        table.extend(devices)
        if table:
            result_tables.append(table)

        return result_tables

    def selected_devices(self, parsed_args, api: Rest) -> List[DeviceInfo]:
        regex = parsed_args.regex or parsed_args.not_regex
        matched_items = [
            DeviceInfo(name, system_ip, site_id, state, d_type, model)
            for _, name, system_ip, site_id, state, d_type, model in Device.get_raise(api).extended_iter(default='-')
            if ((regex is None or regex_search(regex, name, d_type, model, inverse=parsed_args.regex is None)) and
                (not parsed_args.reachable or state == 'reachable') and
                (parsed_args.site is None or site_id == parsed_args.site) and
                (parsed_args.system_ip is None or system_ip == parsed_args.system_ip))
        ]
        # Sort device list by hostname then system ip
        matched_items.sort(key=attrgetter('hostname', 'system_ip'))

        self.log_info(f'Device selection matched {len(matched_items)} devices')

        return matched_items

    def build_table(self, name: str, headers: Sequence[str], devices: Sequence[DeviceInfo], device_data: dict) -> Table:
        table = Table('Device', *headers, name=name)
        for device in devices:
            device_row_values = device_data.get(device.system_ip)
            if device_row_values is None:
                self.log_info(f'{name} missing for {device.hostname}')
                continue

            # Row values are sorted to ensure consistent output
            device_row_values.sort()
            table.extend((device.hostname, *row_values) for row_values in device_row_values)
            table.add_marker()

        return table

    def records(self, parsed_args, api: Rest) -> List[Table]:
        device_map = {system_ip: name for _, name, system_ip, *_ in Device.get_raise(api).extended_iter(default='-')}

        def device_names(devices_field):
            system_ips = (entry.get('system-ip', '') for entry in devices_field)
            return ', '.join(device_map.get(system_ip, system_ip) for system_ip in system_ips)

        end_time = datetime.now(tz=timezone.utc)
        start_time = end_time - timedelta(days=parsed_args.days, hours=parsed_args.hours)
        self.log_info(f'Records query: {start_time:%Y-%m-%d %H:%M:%S %Z} -> {end_time:%Y-%m-%d %H:%M:%S %Z}')

        result_tables = []
        op_obj: RecordItem = parsed_args.subtask_op_cls.get(api, start_time=start_time, end_time=end_time,
                                                            max_records=parsed_args.max)
        if op_obj is None:
            self.log_error(f'Failed to retrieve {parsed_args.subtask_info.lower()}')
        else:
            fields = table_fields(parsed_args.subtask_op_cls, parsed_args.detail, parsed_args.simple)
            field_conversion_fns = {**op_obj.field_conversion_fns, 'devices': device_names}
            table = Table(*op_obj.field_info(*fields))
            table.extend(row for row in op_obj.field_value_iter(*fields, **field_conversion_fns))

            if table:
                result_tables.append(table)

        return result_tables


class ShowArgs(TableTaskArgs):
    subtask_info: str
    subtask_handler: Callable
    regex: Optional[str] = None
    not_regex: Optional[str] = None
    reachable: bool = False
    site: Optional[str] = None
    system_ip: Optional[str] = None

    # Validators
    _validate_regex = validator('regex', 'not_regex', allow_reuse=True)(validate_regex)
    _validate_site_id = validator('site', allow_reuse=True)(validate_site_id)
    _validate_ipv4 = validator('system_ip', allow_reuse=True)(validate_ipv4)

    @root_validator(skip_on_failure=True)
    def mutex_validations(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get('regex') is not None and values.get('not_regex') is not None:
            raise ValueError('Argument "not_regex" not allowed with "regex"')

        if values.get('detail', False) and values.get('simple', False):
            raise ValueError('Argument "detail" not allowed with "simple"')

        return values


class ShowDevicesArgs(ShowArgs):
    subtask_info: str = const('devices')
    subtask_handler: Callable = const(TaskShow.devices)


class ShowRealtimeArgs(ShowArgs):
    subtask_info: str = const('realtime')
    subtask_handler: Callable = const(TaskShow.realtime)
    cmd: List[str]
    detail: bool = False
    simple: bool = False

    # Validators
    @validator('cmd')
    def validate_cmd(cls, cmd_list: List[str]) -> List[str]:
        return validate_op_cmd(OpType.RT, cmd_list)


class ShowStateArgs(ShowArgs):
    subtask_info: str = const('state')
    subtask_handler: Callable = const(TaskShow.bulk_state)
    cmd: List[str]
    detail: bool = False
    simple: bool = False

    # Validators
    @validator('cmd')
    def validate_cmd(cls, cmd_list: List[str]) -> List[str]:
        return validate_op_cmd(OpType.STATE, cmd_list)


class ShowStatisticsArgs(ShowArgs):
    subtask_info: str = const('statistics')
    subtask_handler: Callable = const(TaskShow.bulk_stats)
    cmd: List[str]
    detail: bool = False
    simple: bool = False
    days: conint(ge=0, lt=10000) = 0
    hours: conint(ge=0, lt=10000) = 0

    # Validators
    @validator('cmd')
    def validate_cmd(cls, cmd_list: List[str]) -> List[str]:
        return validate_op_cmd(OpType.STATS, cmd_list)


class ShowRecordsArgs(TableTaskArgs):
    subtask_info: str
    subtask_handler: Callable = const(TaskShow.records)
    subtask_op_cls: Callable
    max: conint(ge=1, lt=1000000) = 100
    days: conint(ge=0, lt=10000) = 0
    hours: conint(ge=0, lt=10000) = 1
    detail: bool = False
    simple: bool = False

    @root_validator(skip_on_failure=True)
    def mutex_validations(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get('detail', False) and values.get('simple', False):
            raise ValueError('Argument "detail" not allowed with "simple"')

        return values


class ShowAlarmsArgs(ShowRecordsArgs):
    subtask_info: str = const('alarms')
    subtask_op_cls: Callable = const(Alarm)


class ShowEventsArgs(ShowRecordsArgs):
    subtask_info: str = const('events')
    subtask_op_cls: Callable = const(Event)
